"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.cliOptionsTypes = exports.shorthands = exports.commandNames = void 0;
exports.rcOptionsTypes = rcOptionsTypes;
exports.help = help;
exports.handler = handler;
exports.createCacheKey = createCacheKey;
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const util_1 = __importDefault(require("util"));
const cli_utils_1 = require("@pnpm/cli-utils");
const common_cli_options_help_1 = require("@pnpm/common-cli-options-help");
const config_1 = require("@pnpm/config");
const crypto_base32_hash_1 = require("@pnpm/crypto.base32-hash");
const error_1 = require("@pnpm/error");
const plugin_commands_installation_1 = require("@pnpm/plugin-commands-installation");
const read_package_json_1 = require("@pnpm/read-package-json");
const package_bins_1 = require("@pnpm/package-bins");
const execa_1 = __importDefault(require("execa"));
const omit_1 = __importDefault(require("ramda/src/omit"));
const pick_1 = __importDefault(require("ramda/src/pick"));
const render_help_1 = __importDefault(require("render-help"));
const symlink_dir_1 = __importDefault(require("symlink-dir"));
const makeEnv_1 = require("./makeEnv");
exports.commandNames = ['dlx'];
exports.shorthands = {
    c: '--shell-mode',
};
function rcOptionsTypes() {
    return {
        ...(0, pick_1.default)([
            'use-node-version',
        ], config_1.types),
        'shell-mode': Boolean,
    };
}
const cliOptionsTypes = () => ({
    ...rcOptionsTypes(),
    package: [String, Array],
});
exports.cliOptionsTypes = cliOptionsTypes;
function help() {
    return (0, render_help_1.default)({
        description: 'Run a package in a temporary environment.',
        descriptionLists: [
            {
                title: 'Options',
                list: [
                    {
                        description: 'The package to install before running the command',
                        name: '--package',
                    },
                    {
                        description: 'Runs the script inside of a shell. Uses /bin/sh on UNIX and \\cmd.exe on Windows.',
                        name: '--shell-mode',
                        shortAlias: '-c',
                    },
                ],
            },
            common_cli_options_help_1.OUTPUT_OPTIONS,
        ],
        url: (0, cli_utils_1.docsUrl)('dlx'),
        usages: ['pnpm dlx <command> [args...]'],
    });
}
async function handler(opts, [command, ...args]) {
    const pkgs = opts.package ?? [command];
    const { cacheLink, cacheExists, cachedDir } = findCache(pkgs, {
        dlxCacheMaxAge: opts.dlxCacheMaxAge,
        cacheDir: opts.cacheDir,
        registries: opts.registries,
    });
    if (!cacheExists) {
        fs_1.default.mkdirSync(cachedDir, { recursive: true });
        await plugin_commands_installation_1.add.handler({
            // Ideally the config reader should ignore these settings when the dlx command is executed.
            // This is a temporary solution until "@pnpm/config" is refactored.
            ...(0, omit_1.default)(['workspaceDir', 'rootProjectManifest', 'symlink'], opts),
            bin: path_1.default.join(cachedDir, 'node_modules/.bin'),
            dir: cachedDir,
            lockfileDir: cachedDir,
            rootProjectManifestDir: cachedDir, // This property won't be used as rootProjectManifest will be undefined
            saveProd: true, // dlx will be looking for the package in the "dependencies" field!
            saveDev: false,
            saveOptional: false,
            savePeer: false,
        }, pkgs);
        try {
            await (0, symlink_dir_1.default)(cachedDir, cacheLink, { overwrite: true });
        }
        catch (error) {
            // EBUSY means that there is another dlx process running in parallel that has acquired the cache link first.
            // Similarly, EEXIST means that another dlx process has created the cache link before this process.
            // The link created by the other process is just as up-to-date as the link the current process was attempting
            // to create. Therefore, instead of re-attempting to create the current link again, it is just as good to let
            // the other link stay. The current process should yield.
            if (!util_1.default.types.isNativeError(error) || !('code' in error) || (error.code !== 'EBUSY' && error.code !== 'EEXIST')) {
                throw error;
            }
        }
    }
    const modulesDir = path_1.default.join(cachedDir, 'node_modules');
    const binsDir = path_1.default.join(modulesDir, '.bin');
    const env = (0, makeEnv_1.makeEnv)({
        userAgent: opts.userAgent,
        prependPaths: [binsDir, ...opts.extraBinPaths],
    });
    const binName = opts.package
        ? command
        : await getBinName(modulesDir, await getPkgName(cachedDir));
    try {
        await (0, execa_1.default)(binName, args, {
            cwd: process.cwd(),
            env,
            stdio: 'inherit',
            shell: opts.shellMode ?? false,
        });
    }
    catch (err) {
        if (util_1.default.types.isNativeError(err) && 'exitCode' in err && err.exitCode != null) {
            return {
                exitCode: err.exitCode,
            };
        }
        throw err;
    }
    return { exitCode: 0 };
}
async function getPkgName(pkgDir) {
    const manifest = await (0, read_package_json_1.readPackageJsonFromDir)(pkgDir);
    const dependencyNames = Object.keys(manifest.dependencies ?? {});
    if (dependencyNames.length === 0) {
        throw new error_1.PnpmError('DLX_NO_DEP', 'dlx was unable to find the installed dependency in "dependencies"');
    }
    return dependencyNames[0];
}
async function getBinName(modulesDir, pkgName) {
    const pkgDir = path_1.default.join(modulesDir, pkgName);
    const manifest = await (0, read_package_json_1.readPackageJsonFromDir)(pkgDir);
    const bins = await (0, package_bins_1.getBinsFromPackageManifest)(manifest, pkgDir);
    if (bins.length === 0) {
        throw new error_1.PnpmError('DLX_NO_BIN', `No binaries found in ${pkgName}`);
    }
    if (bins.length === 1) {
        return bins[0].name;
    }
    const scopelessPkgName = scopeless(manifest.name);
    const defaultBin = bins.find(({ name }) => name === scopelessPkgName);
    if (defaultBin)
        return defaultBin.name;
    const binNames = bins.map(({ name }) => name);
    throw new error_1.PnpmError('DLX_MULTIPLE_BINS', `Could not determine executable to run. ${pkgName} has multiple binaries: ${binNames.join(', ')}`, {
        hint: `Try one of the following:
${binNames.map(name => `pnpm --package=${pkgName} dlx ${name}`).join('\n')}
`,
    });
}
function scopeless(pkgName) {
    if (pkgName[0] === '@') {
        return pkgName.split('/')[1];
    }
    return pkgName;
}
function findCache(pkgs, opts) {
    const dlxCommandCacheDir = createDlxCommandCacheDir(pkgs, opts);
    const cacheLink = path_1.default.join(dlxCommandCacheDir, 'pkg');
    const cachedDir = getValidCacheDir(cacheLink, opts.dlxCacheMaxAge);
    return {
        cacheLink,
        cachedDir: cachedDir ?? getPrepareDir(dlxCommandCacheDir),
        cacheExists: cachedDir != null,
    };
}
function createDlxCommandCacheDir(pkgs, opts) {
    const dlxCacheDir = path_1.default.resolve(opts.cacheDir, 'dlx');
    const cacheKey = createCacheKey(pkgs, opts.registries);
    const cachePath = path_1.default.join(dlxCacheDir, cacheKey);
    fs_1.default.mkdirSync(cachePath, { recursive: true });
    return cachePath;
}
function createCacheKey(pkgs, registries) {
    const sortedPkgs = [...pkgs].sort((a, b) => a.localeCompare(b));
    const sortedRegistries = Object.entries(registries).sort(([k1], [k2]) => k1.localeCompare(k2));
    const hashStr = JSON.stringify([sortedPkgs, sortedRegistries]);
    return (0, crypto_base32_hash_1.createBase32Hash)(hashStr);
}
function getValidCacheDir(cacheLink, dlxCacheMaxAge) {
    let stats;
    let target;
    try {
        stats = fs_1.default.lstatSync(cacheLink);
        if (stats.isSymbolicLink()) {
            target = fs_1.default.realpathSync(cacheLink);
            if (!target)
                return undefined;
        }
        else {
            return undefined;
        }
    }
    catch (err) {
        if (util_1.default.types.isNativeError(err) && 'code' in err && err.code === 'ENOENT') {
            return undefined;
        }
        throw err;
    }
    const isValid = stats.mtime.getTime() + dlxCacheMaxAge * 60_000 >= new Date().getTime();
    return isValid ? target : undefined;
}
function getPrepareDir(cachePath) {
    const name = `${new Date().getTime().toString(16)}-${process.pid.toString(16)}`;
    return path_1.default.join(cachePath, name);
}
//# sourceMappingURL=dlx.js.map