"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.cliOptionsTypes = exports.commandNames = exports.shorthands = void 0;
exports.rcOptionsTypes = rcOptionsTypes;
exports.help = help;
exports.getResumedPackageChunks = getResumedPackageChunks;
exports.writeRecursiveSummary = writeRecursiveSummary;
exports.createEmptyRecursiveSummary = createEmptyRecursiveSummary;
exports.getExecutionDuration = getExecutionDuration;
exports.handler = handler;
const path_1 = __importDefault(require("path"));
const cli_utils_1 = require("@pnpm/cli-utils");
const core_loggers_1 = require("@pnpm/core-loggers");
const common_cli_options_help_1 = require("@pnpm/common-cli-options-help");
const config_1 = require("@pnpm/config");
const lifecycle_1 = require("@pnpm/lifecycle");
const logger_1 = require("@pnpm/logger");
const read_project_manifest_1 = require("@pnpm/read-project-manifest");
const plugin_commands_env_1 = require("@pnpm/plugin-commands-env");
const sort_packages_1 = require("@pnpm/sort-packages");
const execa_1 = __importDefault(require("execa"));
const p_limit_1 = __importDefault(require("p-limit"));
const env_path_1 = require("@pnpm/env.path");
const pick_1 = __importDefault(require("ramda/src/pick"));
const render_help_1 = __importDefault(require("render-help"));
const existsInDir_1 = require("./existsInDir");
const makeEnv_1 = require("./makeEnv");
const run_1 = require("./run");
const error_1 = require("@pnpm/error");
const which_1 = __importDefault(require("which"));
const write_json_file_1 = __importDefault(require("write-json-file"));
const buildCommandNotFoundHint_1 = require("./buildCommandNotFoundHint");
exports.shorthands = {
    parallel: run_1.shorthands.parallel,
    c: '--shell-mode',
};
exports.commandNames = ['exec'];
function rcOptionsTypes() {
    return {
        ...(0, pick_1.default)([
            'bail',
            'sort',
            'use-node-version',
            'unsafe-perm',
            'workspace-concurrency',
            'reporter-hide-prefix',
        ], config_1.types),
        'shell-mode': Boolean,
        'resume-from': String,
        'report-summary': Boolean,
    };
}
const cliOptionsTypes = () => ({
    ...rcOptionsTypes(),
    recursive: Boolean,
    reverse: Boolean,
});
exports.cliOptionsTypes = cliOptionsTypes;
function help() {
    return (0, render_help_1.default)({
        description: 'Run a shell command in the context of a project.',
        descriptionLists: [
            {
                title: 'Options',
                list: [
                    {
                        description: 'Do not hide project name prefix from output of recursively running command.',
                        name: '--no-reporter-hide-prefix',
                    },
                    run_1.PARALLEL_OPTION_HELP,
                    {
                        description: 'Run the shell command in every package found in subdirectories \
or every workspace package, when executed inside a workspace. \
For options that may be used with `-r`, see "pnpm help recursive"',
                        name: '--recursive',
                        shortAlias: '-r',
                    },
                    {
                        description: 'If exist, runs file inside of a shell. \
Uses /bin/sh on UNIX and \\cmd.exe on Windows. \
The shell should understand the -c switch on UNIX or /d /s /c on Windows.',
                        name: '--shell-mode',
                        shortAlias: '-c',
                    },
                    run_1.RESUME_FROM_OPTION_HELP,
                    run_1.REPORT_SUMMARY_OPTION_HELP,
                    ...common_cli_options_help_1.UNIVERSAL_OPTIONS,
                ],
            },
            common_cli_options_help_1.FILTERING,
        ],
        url: (0, cli_utils_1.docsUrl)('exec'),
        usages: ['pnpm [-r] [-c] exec <command> [args...]'],
    });
}
function getResumedPackageChunks({ resumeFrom, chunks, selectedProjectsGraph, }) {
    const resumeFromPackagePrefix = Object.keys(selectedProjectsGraph)
        .find((prefix) => selectedProjectsGraph[prefix]?.package.manifest.name === resumeFrom);
    if (!resumeFromPackagePrefix) {
        throw new error_1.PnpmError('RESUME_FROM_NOT_FOUND', `Cannot find package ${resumeFrom}. Could not determine where to resume from.`);
    }
    const chunkPosition = chunks.findIndex(chunk => chunk.includes(resumeFromPackagePrefix));
    return chunks.slice(chunkPosition);
}
async function writeRecursiveSummary(opts) {
    await (0, write_json_file_1.default)(path_1.default.join(opts.dir, 'pnpm-exec-summary.json'), {
        executionStatus: opts.summary,
    });
}
function createEmptyRecursiveSummary(chunks) {
    const acc = {};
    for (const prefix of chunks.flat()) {
        acc[prefix] = { status: 'queued' };
    }
    return acc;
}
function getExecutionDuration(start) {
    const end = process.hrtime(start);
    return (end[0] * 1e9 + end[1]) / 1e6;
}
async function handler(opts, params) {
    // For backward compatibility
    if (params[0] === '--') {
        params.shift();
    }
    if (!params[0]) {
        throw new error_1.PnpmError('EXEC_MISSING_COMMAND', '\'pnpm exec\' requires a command to run');
    }
    const limitRun = (0, p_limit_1.default)(opts.workspaceConcurrency ?? 4);
    let chunks;
    if (opts.recursive) {
        chunks = opts.sort
            ? (0, sort_packages_1.sortPackages)(opts.selectedProjectsGraph)
            : [Object.keys(opts.selectedProjectsGraph).sort()];
        if (opts.reverse) {
            chunks = chunks.reverse();
        }
    }
    else {
        chunks = [[opts.dir]];
        const project = await (0, read_project_manifest_1.tryReadProjectManifest)(opts.dir);
        if (project.manifest != null) {
            opts.selectedProjectsGraph = {
                [opts.dir]: {
                    dependencies: [],
                    package: {
                        ...project,
                        rootDir: opts.dir,
                        rootDirRealPath: opts.dir,
                    },
                },
            };
        }
    }
    if (!opts.selectedProjectsGraph) {
        throw new error_1.PnpmError('RECURSIVE_EXEC_NO_PACKAGE', 'No package found in this workspace');
    }
    if (opts.resumeFrom) {
        chunks = getResumedPackageChunks({
            resumeFrom: opts.resumeFrom,
            chunks,
            selectedProjectsGraph: opts.selectedProjectsGraph,
        });
    }
    const result = createEmptyRecursiveSummary(chunks);
    const existsPnp = existsInDir_1.existsInDir.bind(null, '.pnp.cjs');
    const workspacePnpPath = opts.workspaceDir && existsPnp(opts.workspaceDir);
    let exitCode = 0;
    const mapPrefixToPrependPaths = {};
    await Promise.all(chunks.flat().map(async (prefix) => {
        const executionEnv = await (0, plugin_commands_env_1.prepareExecutionEnv)(opts, {
            extraBinPaths: opts.extraBinPaths,
            executionEnv: opts.selectedProjectsGraph[prefix]?.package.manifest.pnpm?.executionEnv,
        });
        mapPrefixToPrependPaths[prefix] = [
            './node_modules/.bin',
            ...executionEnv.extraBinPaths,
        ];
    }));
    const reporterShowPrefix = opts.recursive && opts.reporterHidePrefix === false;
    for (const chunk of chunks) {
        // eslint-disable-next-line no-await-in-loop
        await Promise.all(chunk.map(async (prefix) => limitRun(async () => {
            const prependPaths = mapPrefixToPrependPaths[prefix];
            result[prefix].status = 'running';
            const startTime = process.hrtime();
            try {
                const pnpPath = workspacePnpPath ?? existsPnp(prefix);
                const extraEnv = {
                    ...opts.extraEnv,
                    ...(pnpPath ? (0, lifecycle_1.makeNodeRequireOption)(pnpPath) : {}),
                };
                const env = (0, makeEnv_1.makeEnv)({
                    extraEnv: {
                        ...extraEnv,
                        PNPM_PACKAGE_NAME: opts.selectedProjectsGraph[prefix]?.package.manifest.name,
                        ...(opts.nodeOptions ? { NODE_OPTIONS: opts.nodeOptions } : {}),
                    },
                    prependPaths,
                    userAgent: opts.userAgent,
                });
                const [cmd, ...args] = params;
                if (reporterShowPrefix) {
                    const manifest = await (0, cli_utils_1.readProjectManifestOnly)(prefix);
                    const child = (0, execa_1.default)(cmd, args, {
                        cwd: prefix,
                        env,
                        stdio: 'pipe',
                        shell: opts.shellMode ?? false,
                    });
                    const lifecycleOpts = {
                        wd: prefix,
                        depPath: manifest.name ?? path_1.default.relative(opts.dir, prefix),
                        stage: '(exec)',
                    };
                    const logFn = (stdio) => (data) => {
                        for (const line of String(data).split('\n')) {
                            core_loggers_1.lifecycleLogger.debug({
                                ...lifecycleOpts,
                                stdio,
                                line,
                            });
                        }
                    };
                    child.stdout.on('data', logFn('stdout'));
                    child.stderr.on('data', logFn('stderr'));
                    void child.once('close', exitCode => {
                        core_loggers_1.lifecycleLogger.debug({
                            ...lifecycleOpts,
                            exitCode: exitCode ?? 1,
                            optional: false,
                        });
                    });
                    await child;
                }
                else {
                    await (0, execa_1.default)(cmd, args, {
                        cwd: prefix,
                        env,
                        stdio: 'inherit',
                        shell: opts.shellMode ?? false,
                    });
                }
                result[prefix].status = 'passed';
                result[prefix].duration = getExecutionDuration(startTime);
            }
            catch (err) { // eslint-disable-line
                if (isErrorCommandNotFound(params[0], err, prependPaths)) {
                    err.message = `Command "${params[0]}" not found`;
                    err.hint = await createExecCommandNotFoundHint(params[0], {
                        implicitlyFellbackFromRun: opts.implicitlyFellbackFromRun ?? false,
                        dir: opts.dir,
                        workspaceDir: opts.workspaceDir,
                        modulesDir: opts.modulesDir ?? 'node_modules',
                    });
                }
                else if (!opts.recursive && typeof err.exitCode === 'number') {
                    exitCode = err.exitCode;
                    return;
                }
                logger_1.logger.info(err);
                result[prefix] = {
                    status: 'failure',
                    duration: getExecutionDuration(startTime),
                    error: err,
                    message: err.message,
                    prefix,
                };
                if (!opts.bail) {
                    return;
                }
                if (!err['code']?.startsWith('ERR_PNPM_')) {
                    err['code'] = 'ERR_PNPM_RECURSIVE_EXEC_FIRST_FAIL';
                }
                err['prefix'] = prefix;
                opts.reportSummary && await writeRecursiveSummary({
                    dir: opts.lockfileDir ?? opts.dir,
                    summary: result,
                });
                throw err;
            }
        })));
    }
    opts.reportSummary && await writeRecursiveSummary({
        dir: opts.lockfileDir ?? opts.dir,
        summary: result,
    });
    (0, cli_utils_1.throwOnCommandFail)('pnpm recursive exec', result);
    return { exitCode };
}
async function createExecCommandNotFoundHint(programName, opts) {
    if (opts.implicitlyFellbackFromRun) {
        let nearestScript;
        try {
            nearestScript = (0, buildCommandNotFoundHint_1.getNearestScript)(programName, (await (0, cli_utils_1.readProjectManifestOnly)(opts.dir)).scripts);
        }
        catch { }
        if (nearestScript) {
            return `Did you mean "pnpm ${nearestScript}"?`;
        }
        const nearestProgram = (0, buildCommandNotFoundHint_1.getNearestProgram)({
            programName,
            dir: opts.dir,
            workspaceDir: opts.workspaceDir,
            modulesDir: opts.modulesDir,
        });
        if (nearestProgram) {
            return `Did you mean "pnpm ${nearestProgram}"?`;
        }
        return undefined;
    }
    const nearestProgram = (0, buildCommandNotFoundHint_1.getNearestProgram)({
        programName,
        dir: opts.dir,
        workspaceDir: opts.workspaceDir,
        modulesDir: opts.modulesDir,
    });
    if (nearestProgram) {
        return `Did you mean "pnpm exec ${nearestProgram}"?`;
    }
    return undefined;
}
function isErrorCommandNotFound(command, error, prependPaths) {
    // Mac/Linux
    if (process.platform === 'linux' || process.platform === 'darwin') {
        return error.originalMessage === `spawn ${command} ENOENT`;
    }
    // Windows
    if (process.platform === 'win32') {
        const { value: path } = (0, env_path_1.prependDirsToPath)(prependPaths);
        return !which_1.default.sync(command, {
            nothrow: true,
            path,
        });
    }
    return false;
}
//# sourceMappingURL=exec.js.map